"""Tests for triples."""

import tempfile
import unittest
from pathlib import Path

from curies import Triple
from curies.triples import read_triples, write_triples


class TestTriples(unittest.TestCase):
    """Test triples."""

    def test_roundtrip(self) -> None:
        """Test roundtrip."""
        triples = [
            Triple.from_curies("a:1", "a:2", "a:3"),
            Triple.from_curies("a:1", "a:2", "a:4"),
        ]
        with tempfile.TemporaryDirectory() as directory:
            for path in [
                Path(directory).joinpath("test.tsv.gz"),
                Path(directory).joinpath("test.tsv"),
            ]:
                with self.subTest(path=path):
                    write_triples(triples, path)
                    reconstituted = read_triples(path)
                    self.assertEqual(triples, reconstituted)



"""Generate a JSON schema for extended prefix maps."""

import json
from pathlib import Path

from pydantic.json_schema import models_json_schema

from curies import Records

HERE = Path(__file__).parent.resolve()
PATH = HERE.joinpath("schema.json")
TITLE = "Extended Prefix Map"
DESCRIPTION = (
    """\
An extended prefix map is a generalization of a prefix map that
includes synonyms for URI prefixes and CURIE prefixes.
""".strip()
    .replace("\n", " ")
    .replace("  ", " ")
)
URL = "https://w3id.org/biopragmatics/schema/epm.json"


def main() -> None:
    """Generate a JSON schema for extended prefix maps."""
    rv = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": URL,
    }

    _, schema_dict = models_json_schema(
        [(Records, "validation")],
        title=TITLE,
        description=DESCRIPTION,
    )

    rv.update(schema_dict)
    PATH.write_text(json.dumps(rv, indent=2) + "\n")


if __name__ == "__main__":
    main()



"""A simple web service for resolving CURIEs."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from .api import Converter

if TYPE_CHECKING:
    import fastapi
    import flask
    from werkzeug.wrappers import Response

__all__ = [
    "get_fastapi_app",
    "get_fastapi_router",
    "get_flask_app",
    "get_flask_blueprint",
]

#: The code for `Unprocessable Entity <https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/422>`_
FAILURE_CODE = 422


def get_flask_blueprint(converter: Converter, **kwargs: Any) -> flask.Blueprint:
    """Get a blueprint for :class:`flask.Flask`.

    :param converter: A converter
    :param kwargs: Keyword arguments passed through to :class:`flask.Blueprint`

    :returns: A blueprint

    The following is an end-to-end example of using this function to create a small web
    resolver application.

    .. code-block::

        # flask_example.py
        from flask import Flask
        from curies import Converter, get_flask_blueprint, get_obo_converter

        # Create a converter
        converter: Converter = get_obo_converter()

        # Create a blueprint from the converter
        blueprint = get_flask_blueprint(converter)

        # Create the Flask app and mount the router
        app = Flask(__name__)
        app.register_blueprint(blueprint)

        if __name__ == "__main__":
            app.run()

    In the command line, either run your Python file directly, or via with
    :mod:`gunicorn`:

    .. code-block:: shell

        pip install gunicorn
        gunicorn --bind 0.0.0.0:8764 flask_example:app

    Test a request in the Python REPL.

    .. code-block::

        >>> import requests
        >>> requests.get("http://localhost:8764/GO:0032571").url
        'http://amigo.geneontology.org/amigo/term/GO:0032571'
    """
    from flask import Blueprint, abort, redirect

    blueprint = Blueprint("metaresolver", __name__, **kwargs)

    @blueprint.route(f"/<prefix>{converter.delimiter}<path:identifier>")
    def resolve(prefix: str, identifier: str) -> Response:
        """Resolve a CURIE."""
        location = converter.expand_pair(prefix, identifier)
        if location is None:
            prefixes = "".join(f"\n- {p}" for p in sorted(converter.get_prefixes()))
            return abort(FAILURE_CODE, f"Invalid prefix: {prefix}. Use one of:{prefixes}")
        return redirect(location)

    return blueprint


def get_flask_app(
    converter: Converter,
    blueprint_kwargs: Mapping[str, Any] | None = None,
    flask_kwargs: Mapping[str, Any] | None = None,
    register_kwargs: Mapping[str, Any] | None = None,
) -> flask.Flask:
    """Get a Flask app.

    :param converter: A converter
    :param blueprint_kwargs: Keyword arguments passed through to
        :class:`flask.Blueprint`
    :param flask_kwargs: Keyword arguments passed through to :class:`flask.Flask`
    :param register_kwargs: Keyword arguments passed through to
        :meth:`flask.Flask.register_blueprint`

    :returns: A Flask app

    .. seealso::

        This function wraps :func:`get_flask_blueprint`. If you already have your own
        Flask app, :func:`get_flask_blueprint` can be used to create a blueprint that
        you can mount using :meth:`flask.Flask.register_blueprint`.

    The following is an end-to-end example of using this function to create a small web
    resolver application.

    .. code-block::

        # flask_example.py
        from flask import Flask
        from curies import Converter, get_flask_app, get_obo_converter

        # Create a converter
        converter: Converter = get_obo_converter()

        # Create the Flask app
        app: Flask = get_flask_app(converter)

        if __name__ == "__main__":
            app.run()

    In the command line, either run your Python file directly to use Flask/Werkzeug's
    built-in development server, or run it with :mod:`gunicorn`:

    .. code-block:: shell

        pip install gunicorn
        gunicorn --bind 0.0.0.0:8764 flask_example:app

    Alternatively, this package contains a CLI in :mod:`curies.cli` that can be used to
    quickly deploy a resolver based on one of the preset prefix maps, a local prefix
    map, or a remote one via URL. The one-line equivalent of the example file is:

    .. code-block:: shell

        python -m curies --port 8764 --framework flask --server gunicorn obo

    Finally, test a request in the Python REPL.

    .. code-block::

        >>> import requests
        >>> requests.get("http://localhost:8764/GO:0032571").url
        'http://amigo.geneontology.org/amigo/term/GO:0032571'
    """
    from flask import Flask

    blueprint = get_flask_blueprint(converter, **(blueprint_kwargs or {}))
    app = Flask(__name__, **(flask_kwargs or {}))
    app.register_blueprint(blueprint, **(register_kwargs or {}))
    return app


def get_fastapi_router(converter: Converter, **kwargs: Any) -> fastapi.APIRouter:
    """Get a router for :class:`fastapi.FastAPI`.

    :param converter: A converter
    :param kwargs: Keyword arguments passed through to :class:`fastapi.APIRouter`

    :returns: A router

    The following is an end-to-end example of using this function to create a small web
    resolver application.

    Create a python file with your :class:`fastapi.FastAPI` instance:

    .. code-block::

        # fastapi_example.py
        from fastapi import FastAPI
        from curies import Converter, get_fastapi_router

        # Create a converter
        converter = Converter.get_obo_converter()

        # Create a router from the converter
        router = get_fastapi_router(converter)

        # Create the FastAPI and mount the router
        app = FastAPI()
        app.include_router(router)

    In the command line,, run your Python file with :mod:`uvicorn`:

    .. code-block:: shell

        pip install uvicorn
        uvicorn fastapi_example:app --port 8764 --host 0.0.0.0

    Test a request in the Python REPL.

    .. code-block::

        >>> import requests
        >>> requests.get("http://localhost:8764/GO:0032571").url
        'http://amigo.geneontology.org/amigo/term/GO:0032571'
    """
    from fastapi import APIRouter, HTTPException, Path
    from fastapi.responses import RedirectResponse

    api_router = APIRouter(**kwargs)

    @api_router.get(f"/{{prefix}}{converter.delimiter}{{identifier}}")
    def resolve(
        prefix: str = Path(
            title="Prefix",
            description="The Bioregistry prefix corresponding to an identifier resource.",
            examples=["doid"],
        ),
        identifier: str = Path(
            title="Local Unique Identifier",
            description="The local unique identifier in the identifier resource referenced by the prefix.",
        ),
    ) -> RedirectResponse:
        """Resolve a CURIE."""
        location = converter.expand_pair(prefix, identifier)
        if location is None:
            prefixes = ", ".join(sorted(converter.get_prefixes()))
            raise HTTPException(
                status_code=FAILURE_CODE,
                detail=f"Invalid prefix: {prefix}. Use one of: {prefixes}",
            )
        return RedirectResponse(location, status_code=302)

    return api_router


def get_fastapi_app(
    converter: Converter,
    router_kwargs: Mapping[str, Any] | None = None,
    fastapi_kwargs: Mapping[str, Any] | None = None,
    include_kwargs: Mapping[str, Any] | None = None,
) -> fastapi.FastAPI:
    """Get a FastAPI app.

    :param converter: A converter
    :param router_kwargs: Keyword arguments passed through to :class:`fastapi.APIRouter`
    :param fastapi_kwargs: Keyword arguments passed through to :class:`fastapi.FastAPI`
    :param include_kwargs: Keyword arguments passed through to
        :meth:`fastapi.FastAPI.include_router`

    :returns: A FastAPI app

    .. seealso::

        This function wraps :func:`get_fastapi_router`. If you already have your own
        FastAPI app, :func:`get_fastapi_router` can be used to create a
        :class:`fastapi.APIRouter` that you can mount using
        :meth:`fastapi.FastAPI.include_router`.

    The following is an end-to-end example of using this function to create a small web
    resolver application.

    Create a python file with your :class:`fastapi.FastAPI` instance:

    .. code-block::

        # fastapi_example.py
        from fastapi import FastAPI
        from curies import Converter, get_fastapi_app

        # Create a converter
        converter = Converter.get_obo_converter()

        # Create the FastAPI
        app: FastAPI = get_fastapi_app(converter)

    In the command line,, run your Python file with :mod:`uvicorn`:

    .. code-block:: shell

        pip install uvicorn
        uvicorn fastapi_example:app --port 8764 --host 0.0.0.0

    Alternatively, this package contains a CLI in :mod:`curies.cli` that can be used to
    quickly deploy a resolver based on one of the preset prefix maps, a local prefix
    map, or a remote one via URL. The one-line equivalent of the example file is:

    .. code-block:: shell

        python -m curies --framework fastapi --server uvicorn obo

    Finally, test a request in the Python REPL.

    .. code-block::

        >>> import requests
        >>> requests.get("http://localhost:8764/GO:0032571").url
        'http://amigo.geneontology.org/amigo/term/GO:0032571'
    """
    from fastapi import FastAPI

    router = get_fastapi_router(converter, **(router_kwargs or {}))
    app = FastAPI(**(fastapi_kwargs or {}))
    app.include_router(router, **(include_kwargs or {}))
    return app



"""Tests for reconciliation."""

import unittest

from curies import Converter, Record, remap_curie_prefixes, remap_uri_prefixes, rewire
from curies.reconciliation import (
    CycleDetected,
    DuplicateKeys,
    DuplicateValues,
    InconsistentMapping,
    _order_curie_remapping,
)

#: The beginning of URIs used throughout examples
P = "https://example.org"


class TestUtils(unittest.TestCase):
    """Test utilities."""

    def test_ordering(self) -> None:
        """Test ordering."""
        converter = Converter(
            [
                Record(prefix="a", uri_prefix=f"{P}/a/"),
                Record(prefix="b", uri_prefix=f"{P}/b/"),
                Record(prefix="c", uri_prefix=f"{P}/c/"),
            ]
        )
        self.assertEqual(
            [("a", "a1"), ("b", "b1")], _order_curie_remapping(converter, {"a": "a1", "b": "b1"})
        )
        # we want to be as low down the chain first. Test both constructions of the dictionary
        self.assertEqual(
            [("c", "a"), ("b", "c")], _order_curie_remapping(converter, {"c": "a", "b": "c"})
        )
        self.assertEqual(
            [("c", "a"), ("b", "c")], _order_curie_remapping(converter, {"b": "c", "c": "a"})
        )

    def test_duplicate_values(self) -> None:
        """Test detecting bad mapping with duplicate."""
        converter = Converter(
            [
                Record(prefix="a", uri_prefix=f"{P}/a/"),
                Record(prefix="b", uri_prefix=f"{P}/b/"),
                Record(prefix="c", uri_prefix=f"{P}/c/"),
            ]
        )
        curie_remapping = {"b": "c", "a": "c"}
        with self.assertRaises(DuplicateValues):
            _order_curie_remapping(converter, curie_remapping)

    def test_duplicate_keys(self) -> None:
        """Test detecting a bad mapping that contains multiple references to the same record in the keys."""
        converter = Converter(
            [
                Record(prefix="a", prefix_synonyms=["a1"], uri_prefix=f"{P}/a/"),
                Record(prefix="b", uri_prefix=f"{P}/b/"),
                Record(prefix="c", uri_prefix=f"{P}/c/"),
            ]
        )
        curie_remapping = {"a": "c", "a1": "b"}
        with self.assertRaises(DuplicateKeys):
            _order_curie_remapping(converter, curie_remapping)

    def test_duplicate_correspondence(self) -> None:
        """Test detecting a bad mapping containing inconsistent references to the same record in the keys and values."""
        converter = Converter(
            [
                Record(prefix="a", prefix_synonyms=["a1"], uri_prefix=f"{P}/a/"),
                Record(prefix="b", uri_prefix=f"{P}/b/"),
                Record(prefix="c", uri_prefix=f"{P}/c/"),
            ]
        )
        curie_remapping = {"a": "c", "b": "a1"}
        with self.assertRaises(InconsistentMapping):
            _order_curie_remapping(converter, curie_remapping)

    def test_cycles(self) -> None:
        """Test detecting bad mapping with cycles."""
        converter = Converter(
            [
                Record(prefix="a", uri_prefix=f"{P}/a/"),
                Record(prefix="b", uri_prefix=f"{P}/b/"),
                Record(prefix="c", uri_prefix=f"{P}/c/"),
            ]
        )
        curie_remapping = {"b": "c", "c": "b"}
        with self.assertRaises(CycleDetected):
            remap_curie_prefixes(converter, curie_remapping)

        curie_remapping = {"a": "b", "b": "c", "c": "a"}
        with self.assertRaises(CycleDetected):
            _order_curie_remapping(converter, curie_remapping)


class TestCURIERemapping(unittest.TestCase):
    """A test case for CURIE prefix remapping."""

    def test_missing(self) -> None:
        """Test simple upgrade."""
        records = [
            Record(prefix="a", prefix_synonyms=["x"], uri_prefix=f"{P}/a/"),
        ]
        converter = Converter(records)
        curie_remapping = {"b": "c"}
        converter = remap_curie_prefixes(converter, curie_remapping)
        self.assertEqual(records, converter.records)

    def test_simple(self) -> None:
        """Test simple upgrade."""
        records = [
            Record(prefix="a", prefix_synonyms=["x"], uri_prefix=f"{P}/a/"),
        ]
        converter = Converter(records)
        curie_remapping = {"a": "a1"}
        converter = remap_curie_prefixes(converter, curie_remapping)
        self.assertEqual(1, len(converter.records))
        self.assertEqual(
            Record(prefix="a1", prefix_synonyms=["a", "x"], uri_prefix=f"{P}/a/"),
            converter.records[0],
        )

    def test_synonym(self) -> None:
        """Test that an upgrade configuration that would cause a clash does nothing."""
        records = [
            Record(prefix="a", prefix_synonyms=["x"], uri_prefix=f"{P}/a/"),
        ]
        converter = Converter(records)
        curie_remapping = {"a": "x"}
        converter = remap_curie_prefixes(converter, curie_remapping)
        self.assertEqual(1, len(converter.records))
        self.assertEqual(
            Record(prefix="x", prefix_synonyms=["a"], uri_prefix=f"{P}/a/"),
            converter.records[0],
        )

    def test_clash(self) -> None:
        """Test that an upgrade configuration that would cause a clash does nothing."""
        records = [
            Record(prefix="a", prefix_synonyms=["x"], uri_prefix=f"{P}/a/"),
            Record(prefix="b", prefix_synonyms=["y"], uri_prefix=f"{P}/b/"),
        ]
        converter = Converter(records)
        curie_remapping = {"a": "b"}
        converter = remap_curie_prefixes(converter, curie_remapping)
        self.assertEqual(2, len(converter.records))
        self.assertEqual(records, converter.records)

    def test_clash_synonym(self) -> None:
        """Test a clash on a synonym."""
        records = [
            Record(prefix="a", prefix_synonyms=["x"], uri_prefix=f"{P}/a/"),
            Record(prefix="b", prefix_synonyms=["y"], uri_prefix=f"{P}/b/"),
        ]
        converter = Converter(records)
        curie_remapping = {"a": "y"}
        converter = remap_curie_prefixes(converter, curie_remapping)
        self.assertEqual(2, len(converter.records))
        self.assertEqual(records, converter.records)

    def test_simultaneous(self) -> None:
        """Test simultaneous remapping."""
        records = [
            Record(prefix="geo", uri_prefix="https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc="),
            Record(prefix="geogeo", uri_prefix="http://purl.obolibrary.org/obo/GEO_"),
        ]
        converter = Converter(records)
        curie_remapping = {"geo": "ncbi.geo", "geogeo": "geo"}
        converter = remap_curie_prefixes(converter, curie_remapping)
        self.assertEqual(
            [
                Record(
                    prefix="geo",
                    prefix_synonyms=["geogeo"],
                    uri_prefix="http://purl.obolibrary.org/obo/GEO_",
                ),
                Record(
                    prefix="ncbi.geo",
                    uri_prefix="https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=",
                ),
            ],
            converter.records,
        )

    def test_simultaneous_synonym(self) -> None:
        """Test simultaneous remapping with synonyms raises an error."""
        records = [
            Record(
                prefix="geo",
                prefix_synonyms=["ggg"],
                uri_prefix="https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=",
            ),
            Record(prefix="geogeo", uri_prefix="http://purl.obolibrary.org/obo/GEO_"),
        ]
        converter = Converter(records)
        curie_remapping = {"ggg": "ncbi.geo", "geogeo": "geo"}
        with self.assertRaises(InconsistentMapping):
            remap_curie_prefixes(converter, curie_remapping)


class TestURIRemapping(unittest.TestCase):
    """A test case for URI prefix remapping."""

    def test_transitive_error(self) -> None:
        """Test error on transitive remapping."""
        converter = Converter([])
        uri_remapping = {f"{P}/nope/": f"{P}/more-nope/", f"{P}/more-nope/": f"{P}/more-more-nope/"}
        with self.assertRaises(NotImplementedError) as e:
            remap_uri_prefixes(converter, uri_remapping)

        # check that stringification works
        self.assertIn("75", str(e.exception))

    def test_missing(self) -> None:
        """Test simple upgrade."""
        records = [
            Record(prefix="a", uri_prefix=f"{P}/a/"),
        ]
        converter = Converter(records)
        uri_remapping = {f"{P}/nope/": f"{P}/more-nope/"}
        converter = remap_uri_prefixes(converter, uri_remapping)
        self.assertEqual(records, converter.records)

    def test_simple(self) -> None:
        """Test simple upgrade."""
        records = [
            Record(prefix="a", uri_prefix=f"{P}/a/", uri_prefix_synonyms=[f"{P}/a1/"]),
        ]
        converter = Converter(records)
        uri_remapping = {f"{P}/a/": f"{P}/a2/"}
        converter = remap_uri_prefixes(converter, uri_remapping)
        self.assertEqual(1, len(converter.records))
        self.assertEqual(
            Record(
                prefix="a",
                uri_prefix=f"{P}/a2/",
                uri_prefix_synonyms=[f"{P}/a/", f"{P}/a1/"],
            ),
            converter.records[0],
        )

    def test_synonym(self) -> None:
        """Test that an upgrade configuration that would cause a clash does nothing."""
        records = [
            Record(prefix="a", uri_prefix=f"{P}/a/", uri_prefix_synonyms=[f"{P}/a1/"]),
        ]
        converter = Converter(records)
        uri_remapping = {f"{P}/a1/": f"{P}/a2/"}
        converter = remap_uri_prefixes(converter, uri_remapping)
        self.assertEqual(1, len(converter.records))
        self.assertEqual(
            Record(prefix="a", uri_prefix=f"{P}/a2/", uri_prefix_synonyms=[f"{P}/a/", f"{P}/a1/"]),
            converter.records[0],
        )

    def test_clash_preferred(self) -> None:
        """Test that an upgrade configuration that would cause a clash does nothing."""
        records = [
            Record(prefix="a", prefix_synonyms=["x"], uri_prefix=f"{P}/a/"),
            Record(prefix="b", prefix_synonyms=["y"], uri_prefix=f"{P}/b/"),
        ]
        converter = Converter(records)
        upgrades = {f"{P}/a/": f"{P}/b/"}
        converter = remap_uri_prefixes(converter, upgrades)
        self.assertEqual(2, len(converter.records))
        self.assertEqual(records, converter.records)

    def test_clash_synonym(self) -> None:
        """Test clashing with a synonym."""
        records = [
            Record(prefix="a", uri_prefix=f"{P}/a/"),
            Record(prefix="b", uri_prefix=f"{P}/b/", uri_prefix_synonyms=[f"{P}/b1/"]),
        ]
        converter = Converter(records)
        upgrades = {f"{P}/a/": f"{P}/b1/"}
        converter = remap_uri_prefixes(converter, upgrades)
        self.assertEqual(2, len(converter.records))
        self.assertEqual(records, converter.records)


class TestRewire(unittest.TestCase):
    """A test case for rewiring."""

    def test_idempotent(self) -> None:
        """Test that a redundant rewiring doesn't do anything."""
        records = [
            Record(prefix="a", uri_prefix=f"{P}/a/", uri_prefix_synonyms=["https://a.org/"]),
        ]
        converter = Converter(records)
        rewiring = {"a": f"{P}/a/"}
        converter = rewire(converter, rewiring)
        self.assertEqual(records, converter.records)

    def test_upgrade_uri_prefixes_simple(self) -> None:
        """Test simple upgrade."""
        records = [
            Record(prefix="a", uri_prefix=f"{P}/a/", uri_prefix_synonyms=["https://a.org/"]),
        ]
        converter = Converter(records)
        rewiring = {"a": f"{P}/a1/"}
        converter = rewire(converter, rewiring)
        self.assertEqual(1, len(converter.records))
        self.assertEqual(
            Record(
                prefix="a", uri_prefix=f"{P}/a1/", uri_prefix_synonyms=["https://a.org/", f"{P}/a/"]
            ),
            converter.records[0],
        )

    # def test_upgrade_uri_prefixes_add(self):
    #     """Test an upgrade that adds an extra prefix."""
    #     records = [
    #         Record(prefix="a", uri_prefix=f"{P}/a/"),
    #     ]
    #     converter = Converter(records)
    #     rewiring = {"b": f"{P}/b/"}
    #     converter = rewire(converter, rewiring)
    #     self.assertEqual(2, len(converter.records))
    #     self.assertEqual(
    #         [
    #             Record(prefix="a", uri_prefix=f"{P}/a/"),
    #             Record(prefix="b", uri_prefix=f"{P}/b/"),
    #         ],
    #         converter.records,
    #     )

    def test_upgrade_uri_prefixes_clash(self) -> None:
        """Test an upgrade that does nothing since it would create a clash."""
        records = [
            Record(prefix="a", uri_prefix=f"{P}/a/"),
            Record(prefix="b", uri_prefix=f"{P}/b/"),
        ]
        converter = Converter(records)
        rewiring = {"b": f"{P}/a/"}
        converter = rewire(converter, rewiring)
        self.assertEqual(2, len(converter.records))
        self.assertEqual(
            [
                Record(prefix="a", uri_prefix=f"{P}/a/"),
                Record(prefix="b", uri_prefix=f"{P}/b/"),
            ],
            converter.records,
        )

    def test_upgrade_uri_upgrade(self) -> None:
        """Test an upgrade of an existing URI prefix synonym."""
        records = [
            Record(prefix="a", uri_prefix=f"{P}/a/", uri_prefix_synonyms=[f"{P}/a1/"]),
        ]
        converter = Converter(records)
        rewiring = {"a": f"{P}/a1/"}
        converter = rewire(converter, rewiring)
        self.assertEqual(1, len(converter.records))
        self.assertEqual(
            [
                Record(
                    prefix="a",
                    uri_prefix=f"{P}/a1/",
                    uri_prefix_synonyms=[f"{P}/a/"],
                ),
            ],
            converter.records,
        )

    def test_upgrade_uri_upgrade_with_curie_prefix(self) -> None:
        """Test an upgrade of an existing URI prefix synonym via a CURIE prefix synonym."""
        records = [
            Record(
                prefix="a",
                prefix_synonyms=["a1"],
                uri_prefix=f"{P}/a/",
                uri_prefix_synonyms=[f"{P}/a1/"],
            ),
        ]
        converter = Converter(records)
        rewiring = {"a1": f"{P}/a1/"}
        converter = rewire(converter, rewiring)
        self.assertEqual(1, len(converter.records))
        self.assertEqual(
            [
                Record(
                    prefix="a",
                    prefix_synonyms=["a1"],
                    uri_prefix=f"{P}/a1/",
                    uri_prefix_synonyms=[f"{P}/a/"],
                ),
            ],
            converter.records,
        )



"""Reconciliation."""

from __future__ import annotations

import logging
from collections import Counter, defaultdict
from collections.abc import Collection, Mapping

from .api import Converter, Record

__all__ = [
    "remap_curie_prefixes",
    "remap_uri_prefixes",
    "rewire",
]

logger = logging.getLogger(__name__)


class TransitiveError(NotImplementedError):
    """An error when transitive mappings appear."""

    def __init__(self, intersection: Collection[str]) -> None:
        """Initialize the exception.

        :param intersection: The strings that appeared both as keys and values in a
            remapping dictionary (either for CURIEs or URIs)
        """
        self.intersection = intersection

    def __str__(self) -> str:
        return (
            f"Transitive mapping has not been implemented. This is being thrown because "
            f"the following appear in both the keys and values of the remapping: {self.intersection}."
            "\n\nSee discussion at https://github.com/cthoyt/curies/issues/75."
        )


def remap_curie_prefixes(converter: Converter, remapping: Mapping[str, str]) -> Converter:
    """Apply CURIE prefix remappings.

    :param converter: A converter
    :param remapping: A mapping from CURIE prefixes to new CURIE prefixes. Old CURIE
        prefixes become synonyms in the records (i.e., they aren't forgotten).

    :returns: An upgraded converter
    """
    ordering = _order_curie_remapping(converter, remapping)
    intersection = set(remapping).intersection(remapping.values())
    records = {r.prefix: r for r in converter.records}

    modified_records = []
    for old, new_prefix in ordering:
        _old = converter.synonym_to_prefix.get(old)
        if _old is None:
            logger.debug(
                "Remapping %s->%s can not be applied because %s does not appear in the converter. Skipping.",
                old,
                new_prefix,
                old,
            )
            continue

        record = records.pop(_old)
        new_record = converter.get_record(new_prefix)
        if new_record is not None and record != new_record:
            logger.debug(
                "Remapping %s->%s would create a clash because of the existing record %r. Skipping.",
                old,
                new_prefix,
                new_record,
            )
        elif old in intersection:
            record.prefix_synonyms = sorted(
                set(record.prefix_synonyms).difference({old, new_prefix})
            )
            record.prefix = new_prefix
        else:
            record.prefix_synonyms = sorted(
                set(record.prefix_synonyms).union({record.prefix}).difference({new_prefix})
            )
            record.prefix = new_prefix
        modified_records.append(record)

    return Converter([*records.values(), *modified_records])


def remap_uri_prefixes(converter: Converter, remapping: Mapping[str, str]) -> Converter:
    """Apply URI prefix remappings.

    :param converter: A converter
    :param remapping: A mapping from URI prefixes to new URI prefixes. Old URI prefixes
        become synonyms in the records (i.e., they aren't forgotten)

    :returns: An upgraded converter

    :raises TransitiveError: If there are any strings that appear in both the key and
        values of the remapping
    """
    intersection = set(remapping).intersection(remapping.values())
    if intersection:
        raise TransitiveError(intersection)

    records = []
    for record in converter.records:
        new_uri_prefix = _get_uri_preferred_or_synonym(record, remapping)
        if new_uri_prefix is None:
            pass  # nothing to upgrade
        elif (
            new_uri_prefix in converter.reverse_prefix_map
            and new_uri_prefix not in record.uri_prefix_synonyms
        ):
            pass  # would create a clash, don't do anything
        else:
            record.uri_prefix_synonyms = sorted(
                set(record.uri_prefix_synonyms)
                .union({record.uri_prefix})
                .difference({new_uri_prefix})
            )
            record.uri_prefix = new_uri_prefix
        records.append(record)
    return Converter(records)


def rewire(converter: Converter, rewiring: Mapping[str, str]) -> Converter:
    """Apply URI prefix upgrades.

    :param converter: A converter
    :param rewiring: A mapping from CURIE prefixes to new URI prefixes. If CURIE
        prefixes are not already in the converter, new records are created. If new URI
        prefixes clash with any existing ones, they are not added.

    :returns: An upgraded converter
    """
    records = []
    for record in converter.records:
        new_uri_prefix = _get_curie_preferred_or_synonym(record, rewiring)
        if new_uri_prefix is None:
            pass  # nothing to upgrade
        elif new_uri_prefix == record.uri_prefix:
            pass  # it's already the preferred prefix, nothing to do
        elif (
            new_uri_prefix in converter.reverse_prefix_map
            and new_uri_prefix not in record.uri_prefix_synonyms
        ):
            logger.debug(
                "Rewiring %r to %s would create a clash because of the existing record %s. Skipping.",
                record,
                new_uri_prefix,
                converter.reverse_prefix_map[new_uri_prefix],
            )
        else:
            record.uri_prefix_synonyms = sorted(
                set(record.uri_prefix_synonyms)
                .union({record.uri_prefix})
                .difference({new_uri_prefix})
            )
            record.uri_prefix = new_uri_prefix
        records.append(record)

    # potential future functionality: add missing records
    # for prefix, new_uri_prefix in rewiring.items():
    #     if prefix not in converter.synonym_to_prefix:
    #         records.append(Record(prefix=prefix, uri_prefix=new_uri_prefix))

    return Converter(records)


def _get_curie_preferred_or_synonym(record: Record, upgrades: Mapping[str, str]) -> str | None:
    if record.prefix in upgrades:
        return upgrades[record.prefix]
    for s in record.prefix_synonyms:
        if s in upgrades:
            return upgrades[s]
    return None


def _get_uri_preferred_or_synonym(record: Record, upgrades: Mapping[str, str]) -> str | None:
    if record.uri_prefix in upgrades:
        return upgrades[record.uri_prefix]
    for s in record.uri_prefix_synonyms:
        if s in upgrades:
            return upgrades[s]
    return None


class DuplicateValues(ValueError):
    """Raised when multiple values in the remapping correspond to the same preferred CURIE prefix."""


class DuplicateKeys(ValueError):
    """Raised when multiple keys in the remapping correspond to the same preferred CURIE prefix."""


class InconsistentMapping(ValueError):
    """Raised when inconsistent prefixes are used in the keys and values of the remapping."""


class CycleDetected(ValueError):
    """Raised when the remapping induces a cycle."""


def _order_curie_remapping(
    converter: Converter, curie_remapping: Mapping[str, str]
) -> list[tuple[str, str]]:
    # Check that no keys of the remapping actually correspond to the same primary prefix
    key_counter = defaultdict(list)
    for key in curie_remapping:
        key_counter[converter.standardize_prefix(key)].append(key)
    duplicate_keys = {
        k: Counter(values) for k, values in key_counter.items() if len(values) > 1 and k is not None
    }
    if duplicate_keys:
        raise DuplicateKeys(f"Duplicate keys in remapping: {duplicate_keys}")

    # Check that it's not the case that multiple prefixes are mapping
    # to the same new prefix.
    value_counter = defaultdict(list)
    for value in curie_remapping.values():
        value_counter[converter.standardize_prefix(value)].append(value)
    duplicate_values = {
        k: Counter(values)
        for k, values in value_counter.items()
        if len(values) > 1 and k is not None
    }
    if duplicate_values:
        raise DuplicateValues(f"Duplicate values in remapping: {duplicate_values}")

    # Check that the correspondence is same for both
    correspondence_counter = defaultdict(set)
    for key, value in curie_remapping.items():
        norm_key = converter.standardize_prefix(key)
        norm_val = converter.standardize_prefix(value)
        correspondence_counter[norm_key].add(key)
        # don't penalize synonym remappings
        if norm_key != norm_val:
            correspondence_counter[norm_val].add(value)
    duplicate_correspondence = {
        k: Counter(values)
        for k, values in correspondence_counter.items()
        if len(values) > 1 and k is not None
    }
    if duplicate_correspondence:
        raise InconsistentMapping(
            f"Inconsistent usage of prefixes in keys and values: {duplicate_correspondence}"
        )

    # Given the two tests before, we don't have to worry about any clashes, and
    # we can work directly on primary prefixes
    if not set(curie_remapping).intersection(curie_remapping.values()):
        # No logic necessary, so just sort based on key to be consistent
        return sorted(curie_remapping.items())

    # assume that there are no duplicates in the values
    rv = []
    d = dict(curie_remapping)
    while d:
        no_outgoing = set(d.values()).difference(d)
        if not no_outgoing:
            raise CycleDetected("cycle detected in remapping")
        edges = sorted((k, v) for k, v in d.items() if v in no_outgoing)
        rv.extend(edges)
        d = {k: v for k, v in d.items() if v not in no_outgoing}
    return rv



"""Utilities for triples."""

from __future__ import annotations

import csv
import gzip
from collections.abc import Generator, Iterable
from contextlib import contextmanager
from pathlib import Path
from typing import NamedTuple, TextIO

from typing_extensions import Self

from curies import Reference

__all__ = [
    "Triple",
    "read_triples",
    "write_triples",
]


class Triple(NamedTuple):
    """A three-tuple of reference, useful for semantic web applications."""

    subject: Reference
    predicate: Reference
    object: Reference

    @classmethod
    def from_curies(cls, subject_curie: str, predicate_curie: str, object_curie: str) -> Self:
        """Construct a triple from three CURIE strings."""
        return cls(
            Reference.from_curie(subject_curie),
            Reference.from_curie(predicate_curie),
            Reference.from_curie(object_curie),
        )


HEADER = Triple._fields


@contextmanager
def _get_file(path: str | Path, read: bool) -> Generator[TextIO, None, None]:
    path = Path(path).expanduser().resolve()
    if path.suffix == ".gz":
        yield gzip.open(path, mode="rt" if read else "wt")
    else:
        yield open(path, mode="r" if read else "w")


def write_triples(triples: Iterable[Triple], path: str | Path) -> None:
    """Write triples to a file."""
    with _get_file(path, read=False) as file:
        writer = csv.writer(file, delimiter="\t")
        writer.writerow(HEADER)
        writer.writerows(
            (triple.subject.curie, triple.predicate.curie, triple.object.curie)
            for triple in triples
        )


def read_triples(path: str | Path, *, reference_cls: type[Reference] | None = None) -> list[Triple]:
    """Read triples."""
    if reference_cls is None:
        reference_cls = Reference
    with _get_file(path, read=True) as file:
        reader = csv.reader(file, delimiter="\t")
        _header = next(reader)
        return [
            Triple(
                reference_cls.from_curie(subject_curie),
                reference_cls.from_curie(predicate_curie),
                reference_cls.from_curie(object_curie),
            )
            for subject_curie, predicate_curie, object_curie in reader
        ]



"""Test writing I/O."""

import json
import unittest
from collections.abc import Iterable
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import cast

import rdflib

import curies
from curies import Converter

CHEBI_URI_PREFIX = "http://purl.obolibrary.org/obo/CHEBI_"


class TestIO(unittest.TestCase):
    """Test I/O."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.prefix = "CHEBI"
        self.uri_prefix = CHEBI_URI_PREFIX
        self.prefix_synonym = "p"
        self.uri_prefix_synonym = "u"
        self.pattern = "^\\d{7}$"
        self.converter = Converter.from_extended_prefix_map(
            [
                {
                    "prefix": self.prefix,
                    "prefix_synonyms": [self.prefix_synonym],
                    "uri_prefix": self.uri_prefix,
                    "uri_prefix_synonyms": [self.uri_prefix_synonym],
                    "pattern": self.pattern,
                },
            ]
        )

    def test_write_epm(self) -> None:
        """Test writing and reading an extended prefix map."""
        with TemporaryDirectory() as d:
            path = Path(d).joinpath("test.json")
            curies.write_extended_prefix_map(self.converter, path)
            nc = curies.load_extended_prefix_map(path)
        self.assertEqual(self.converter.records, nc.records)
        self.assertEqual({self.prefix: self.pattern}, nc.pattern_map)

    def test_write_jsonld_with_bimap(self) -> None:
        """Test writing and reading a prefix map via JSON-LD."""
        with TemporaryDirectory() as d:
            path = Path(d).joinpath("test.json")
            curies.write_jsonld_context(self.converter, path.as_posix())
            nc = curies.load_jsonld_context(path)
        self.assertEqual({self.prefix: self.uri_prefix}, nc.prefix_map)
        self.assertEqual(
            {self.uri_prefix: self.prefix},
            nc.reverse_prefix_map,
            msg="the prefix synonym should not survive round trip",
        )
        self.assertEqual({self.prefix: self.uri_prefix}, nc.bimap)

    def test_write_jsonld_with_synonyms(self) -> None:
        """Test writing a JSON-LD with synonyms."""
        # note: we don't test loading since loading a JSON-LD with synonyms is undefined
        for expand in [True, False]:
            with self.subTest(expand=expand):
                with TemporaryDirectory() as d:
                    path = Path(d).joinpath("test.json")
                    curies.write_jsonld_context(self.converter, path, include_synonyms=True)
                    data = json.loads(path.read_text())["@context"]
                self.assertEqual({self.prefix, self.prefix_synonym}, set(data))

    def test_shacl(self) -> None:
        """Test round-tripping SHACL."""
        with TemporaryDirectory() as d:
            path = Path(d).joinpath("test.ttl")
            curies.write_shacl(self.converter, path)
            nc = curies.load_shacl(path)
        self.assertEqual(self.converter.bimap, nc.bimap)
        self.assertEqual({self.prefix: self.pattern}, nc.pattern_map)

    def test_shacl_with_synonyms(self) -> None:
        """Test writing SHACL with synonyms."""
        # note: we don't test loading since loading SHACL with synonyms is undefined
        with TemporaryDirectory() as d:
            path = Path(d).joinpath("test.ttl")
            curies.write_shacl(self.converter, path, include_synonyms=True)
            graph = rdflib.Graph()
            graph.parse(location=path.as_posix(), format="turtle")

        query = """\
            SELECT ?prefix
            WHERE {
                ?bnode sh:declare ?declaration .
                ?declaration sh:prefix ?prefix .
            }
        """
        results = cast(Iterable[tuple[str]], graph.query(query))
        self.assertEqual({self.prefix, self.prefix_synonym}, {str(prefix) for (prefix,) in results})



"""Implementation of mapping service."""

from __future__ import annotations

import itertools as itt
from collections.abc import Collection, Iterable
from typing import TYPE_CHECKING, Any, cast

from rdflib import OWL, Graph, URIRef
from rdflib.term import _is_valid_uri

from .rdflib_custom import MappingServiceSPARQLProcessor  # type: ignore
from .utils import CONTENT_TYPE_TO_RDFLIB_FORMAT, handle_header
from ..api import Converter

if TYPE_CHECKING:
    import fastapi
    import flask


def _prepare_predicates(predicates: None | str | Collection[str] = None) -> set[URIRef]:
    if predicates is None:
        return {OWL.sameAs}
    if isinstance(predicates, str):
        return {URIRef(predicates)}
    return {URIRef(predicate) for predicate in predicates}


class MappingServiceGraph(Graph):
    """A service that implements identifier mapping based on a converter."""

    converter: Converter
    query_predicates: set[URIRef]

    def __init__(
        self,
        *args: Any,
        converter: Converter,
        predicates: None | str | list[str] = None,
        **kwargs: Any,
    ) -> None:
        """Instantiate the graph.

        :param args: Positional arguments to pass to :meth:`rdflib.Graph.__init__`
        :param converter: A converter object
        :param predicates: A predicate or set of predicates. If not given, this service
            will use `owl:sameAs` as a predicate for mapping IRIs.
        :param kwargs: Keyword arguments to pass to :meth:`rdflib.Graph.__init__`

        In the following example, a service graph is instantiated using a small example
        converter, then an example SPARQL query is made directly to show how it makes
        results:

        .. code-block:: python

            from curies import Converter
            from curies.mapping_service import CURIEServiceGraph

            converter = Converter.from_priority_prefix_map(
                {
                    "CHEBI": [
                        "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=",
                        "http://identifiers.org/chebi/",
                        "http://purl.obolibrary.org/obo/CHEBI_",
                    ],
                    "GO": ["http://purl.obolibrary.org/obo/GO_"],
                    "OBO": ["http://purl.obolibrary.org/obo/"],
                    ...,
                }
            )
            graph = MappingServiceGraph(converter=converter)

            res = graph.query('''
                SELECT ?o WHERE {
                    VALUES ?s {
                        <http://purl.obolibrary.org/obo/CHEBI_1>
                    }
                    ?s owl:sameAs ?o
                }
            ''')

        The results of this are:

        ====================================== =================================================
        subject                                object
        http://purl.obolibrary.org/obo/CHEBI_1 http://purl.obolibrary.org/obo/CHEBI_1
        http://purl.obolibrary.org/obo/CHEBI_1 http://identifiers.org/chebi/1
        http://purl.obolibrary.org/obo/CHEBI_1 https://www.ebi.ac.uk/chebi/searchId.do?chebiId=1
        ====================================== =================================================
        """
        self.converter = converter
        self.query_predicates = _prepare_predicates(predicates)
        super().__init__(*args, **kwargs)

    def _expand_pair_all(self, uri_in: str) -> list[URIRef]:
        prefix, identifier = self.converter.parse_uri(uri_in)
        if prefix is None or identifier is None:
            return []
        uris = cast(Collection[str], self.converter.expand_pair_all(prefix, identifier))
        # do _is_valid_uri check because some configurations e.g. from Bioregistry might
        # produce invalid URIs e.g., containing spaces
        return [URIRef(uri) for uri in uris if _is_valid_uri(uri)]

    def triples(  # type:ignore
        self, triple: tuple[URIRef, URIRef, URIRef]
    ) -> Iterable[tuple[URIRef, URIRef, URIRef]]:
        """Generate triples, overriden to dynamically generate mappings based on this graph's converter."""
        subj_query, pred_query, obj_query = triple
        if pred_query in self.query_predicates:
            if subj_query is None and obj_query is not None:
                subjects = self._expand_pair_all(obj_query)
                for subj, pred in itt.product(subjects, self.query_predicates):
                    yield subj, pred, obj_query
            elif subj_query is not None and obj_query is None:
                objects = self._expand_pair_all(subj_query)
                for obj, pred in itt.product(objects, self.query_predicates):
                    yield subj_query, pred, obj


def get_flask_mapping_blueprint(
    converter: Converter, route: str = "/sparql", **kwargs: Any
) -> flask.Blueprint:
    """Get a blueprint for :class:`flask.Flask`.

    :param converter: A converter
    :param route: The route of the SPARQL service (relative to the base of the
        Blueprint)
    :param kwargs: Keyword arguments passed through to :class:`flask.Blueprint`

    :returns: A blueprint
    """
    from flask import Blueprint, Response, request

    blueprint = Blueprint("mapping", __name__, **kwargs)
    graph = MappingServiceGraph(converter=converter)
    processor = MappingServiceSPARQLProcessor(graph=graph)

    @blueprint.route(route, methods=["GET", "POST"])
    def serve_sparql() -> Response:
        """Run a SPARQL query and serve the results."""
        sparql = request.values.get("query")
        if not sparql:
            return Response(
                "Missing query (either in args for GET requests, or in form for POST requests)", 400
            )
        content_type = handle_header(request.headers.get("accept"))
        results = graph.query(sparql, processor=processor)
        response = results.serialize(format=CONTENT_TYPE_TO_RDFLIB_FORMAT[content_type])
        return Response(response, content_type=content_type)

    return blueprint


def get_fastapi_router(
    converter: Converter, route: str = "/sparql", **kwargs: Any
) -> fastapi.APIRouter:
    """Get a router for :class:`fastapi.FastAPI`.

    :param converter: A converter
    :param route: The route of the SPARQL service (relative to the base of the API
        router)
    :param kwargs: Keyword arguments passed through to :class:`fastapi.APIRouter`

    :returns: A router
    """
    from fastapi import APIRouter, Form, Header, Query, Response

    api_router = APIRouter(**kwargs)
    graph = MappingServiceGraph(converter=converter)
    processor = MappingServiceSPARQLProcessor(graph=graph)

    def _resolve(accept: str, sparql: str) -> Response:
        content_type = handle_header(accept)
        results = graph.query(sparql, processor=processor)
        response = results.serialize(format=CONTENT_TYPE_TO_RDFLIB_FORMAT[content_type])
        return Response(response, media_type=content_type)

    @api_router.get(route)
    def resolve_get(
        query: str = Query(description="The SPARQL query to run"),
        accept: str = Header(),
    ) -> Response:
        """Run a SPARQL query and serve the results."""
        return _resolve(accept, query)

    @api_router.post(route)
    def resolve_post(
        query: str = Form(description="The SPARQL query to run"),
        accept: str = Header(),
    ) -> Response:
        """Run a SPARQL query and serve the results."""
        return _resolve(accept, query)

    return api_router


def get_flask_mapping_app(converter: Converter) -> flask.Flask:
    """Get a Flask app for the mapping service."""
    from flask import Flask

    blueprint = get_flask_mapping_blueprint(converter)
    app = Flask(__name__)
    app.register_blueprint(blueprint)
    return app


def get_fastapi_mapping_app(converter: Converter) -> fastapi.FastAPI:
    """Get a FastAPI app.

    :param converter: A converter

    :returns: A FastAPI app
    """
    from fastapi import FastAPI

    router = get_fastapi_router(converter)
    app = FastAPI()
    app.include_router(router)
    return app
